﻿/*
 * * LEGAL DISCLAIMER *

The currency markets can do ANYTHING at ANY TIME.
No one can guarantee or forecast how these results will perform or behave in future markets.
Anyone who uses this product or this information is responsible for deciding If, Where, When and How this product and information are used.
Anyone who uses this product or this information is responsible and liable for any outcomes that might result from the use of this product or this information.
There is no warranty or guarantee provided or implied for this product or this information for any purpose.
 */

using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Media;
using Alveo.Interfaces.UserCode;
using Alveo.UserCode;
using Alveo.Common;
using Alveo.Common.Classes;
using Alveo.Common.Enums;

namespace Alveo.UserCode
{
    [Serializable]
    [Description("")]
    public class HLHB_Trend : ExpertAdvisorBase
    {
        string version = "1.05";
        datetime datetime0 = 0;             // minimum datetime
        const string pair = "EUR/USD";      // default curency
        bool startSession;                  // start of session flag
        EMAobj ema1;
        EMAobj ema2;
        iRSIobj rsi;
        int countBars;

        #region Properties      // EA User Settings

        [Category("Settings")]
        [Description("EMA1 Period in Bars [ex: 5]")]
        public int EMA1_period { get; set; }

        [Category("Settings")]
        [Description("EMA1 Period in Bars [ex: 10]")]
        public int EMA2_period { get; set; }

        [Category("Settings")]
        [Description("Lot size for every order in Standard lots. [ex: 0.01]")]
        public double Lotsize { get; set; }

        [Category("Settings")]
        [Description("StopLoss limit for all trades in Pips. [ex: 150]")]
        public int StopLoss { get; set; }

        [Category("Settings")]
        [Description("Trailing StopLoss Enable boolean. [ex: true=on]")]
        public bool TrailingSL { get; set; }

        [Category("Settings")]
        [Description("Takeprofit limit for all trades in Pips. [ex: 400]")]
        public int TakeProfit { get; set; }

        [Category("Settings")]
        [Description("RSI Period in Bars. [ex: 10]")]
        public int RSI_period { get; set; }
        #endregion

        // EA variable declarations
        public bool simulate;               // simulation mode flag
        public bool optimize;               // optimization mode flag
        public string simSymbol;            // symbol for simulation
        public double simAccountBalance;    // Account Balance for Simulation
        public DateTime simTime;            // UTC Time for simulation
        public DateTime simLocalTime;       // Local Time for simulation
        internal TimeFrame simTimeframe;    // chart timeframe for simulation
        public int simBars;                 // number of chart bars for simulation
        Random rnd = new Random(2874);      // random selector for Sl or TP simulation selector

        static HLHB_Trend ea;
        public string strategy;             // strategy name;
        string symbol;                      // chart symbol
        public int curBars;                 // current number of chart bars
        double pipPos;                      // Pip position
        double point;                       // Point multipler
        int ticketNum;                      // order ticketNum
        internal double curPrice;           // current Price
        internal double midPrice;           // middle Price    HL/2
        DateTime curTime;                   // current UTC Time
        internal BarData curBar;            // current Bar
        internal EA_State s;                // EA State Class object
        internal bool riskLimitReached;     // riskLimitReached
        internal bool tradingclosed;        // tradingclosed
        internal bool paused;               // trading Paused, no new trades
        internal Object TLock;              // Lock for TradeLog file access
        internal double accountBalance;     // account balance
        internal double riskLimit;          // in Pips,  i.e. 1.8% of AccountBallance for 1 Standard lot
        string magicStr;

        int total;
        bool initialized;
        bool clearFile;
        int entrySignal;

        internal string dataFileDir;       //directory for data files
        static System.IO.StreamWriter tradeLogFile;             // StreamWriter for tradeLogFile
        static System.IO.StreamWriter debugLogFile;             // StreamWriter for tradeLogFile

        static TimeSpan fridayPause = new TimeSpan(12 + 5, 00, 00);         // Local time
        static TimeSpan fridayclose = new TimeSpan(12 + 4, 50, 00);         // Local time
        static TimeSpan dailyMaintStart = new TimeSpan(12 + 4, 50, 00);     // Local time
        static TimeSpan dailyMaintEnd = new TimeSpan(12 + 5, 15, 00);       // Local time
        static TimeSpan sundayOpen = new TimeSpan(22, 00, 00);              // UTC time

        static int sessionSundayOpen = 0;     // 00:00
        static int sessionSundayClose = 86400; // 24:00
        static int sessionMondayThursdayOpen = 0;     // 00:00
        static int sessionMondayThursdayClose = 86400; // 24:00
        static int sessionFridayOpen = 0;     // 00:00
        static int sessionFridayClose = 86400; // 24:00
        bool sessionIgnoreSunday = true;
        bool sessionCloseAtSessionClose = true;
        bool sessionCloseAtFridayClose = true;

        public bool TPflag;
        public bool SLflag;

        public HLHB_Trend()      // constuctor called by Alveo
        {
            ea = this;
            simulate = false;
            optimize = false;
            InitEA();
        }

        public HLHB_Trend(bool optimizing, DateTime time, string path) : this()  // constructor called by simulation
        {
            simulate = true;
            optimize = optimizing;
            simTime = time;
            InitEA();
            dataFileDir = path;
        }

        public void InitEA()  // Initiatization of EA variables on creation
        {
            // Initializa EA variables
            copyright = "(C)2018 Entity3 LLC. all rights reserved";
            link = "";
            strategy = "HLHB_Trend";
            simSymbol = pair;
            simBars = 0;
            symbol = simSymbol;
            simTimeframe = TimeFrame.M5;
            ticketNum = 0;
            clearFile = true;
            countBars = 0;

            // default trading parameters 
            EMA1_period = 5;
            EMA2_period = 10;
            Lotsize = 0.01;
            StopLoss = 150;
            TrailingSL = true;
            TakeProfit = 400;
            RSI_period = 10;

            curPrice = double.MinValue;
            curTime = GetCurTime();
            curBar = null;
            curBars = 0;
            riskLimit = 1.8;  // in Pips,  i.e. 1.8% of AccountBallance for 1 Standard lot
            riskLimitReached = false;
            simAccountBalance = 1000;
            riskLimitReached = false;
            tradingclosed = false;
            paused = false;
            TLock = new Object();
            s = new EA_State(this);         // create EA State object
            initialized = true;
            ema1 = new EMAobj();
            ema2 = new EMAobj();
            rsi = new iRSIobj();
        }

        //external interface for Simulator
        internal int doInit()
        {
            return this.Init();
        }

        //+------------------------------------------------------------------+"
        //| EA initialization function                                       |"
        //|                                                                  |"
        //| called by Alveo to init EA                                       |"
        //+------------------------------------------------------------------+"
        protected override int Init()
        {
            try
            {
                startSession = true;  // start new session
                paused = false;
                symbol = GetSymbol().Trim();
                var digits = GetDigits();
                pipPos = Math.Pow(10, digits - 1);
                point = 1 / pipPos / 10;
                if (s == null)
                    s = new EA_State(this);
                s.symbol = symbol;
                var tf = GetTimeframe();
                var tfStr = (tf != TimeFrame.Unknown) ? tf.ToString() : "S1";
                dataFileDir = "C:\\temp\\" + strategy + " " + symbol.Replace('/', '-') + " " + tfStr + "\\";       //directory for data files
                magicStr = strategy + "," + symbol.Replace("/", "") + ", " + tfStr;
                if (s.firstRun)
                {
                    if (!optimize)
                    {
                        if (!System.IO.Directory.Exists("C:\\temp"))
                            System.IO.Directory.CreateDirectory("C:\\temp");
                        if (!System.IO.Directory.Exists(dataFileDir))
                            System.IO.Directory.CreateDirectory(dataFileDir);
                    }
                    startSession = true;
                    s.firstRun = false;
                }
                s.ClearState();             // clear EA State object
                if (optimize) // dfb
                    LoadBestParameters();  // load trading parameters from file
                ema1 = new EMAobj(EMA1_period);
                ema2 = new EMAobj(EMA2_period);
                rsi = new iRSIobj(RSI_period);
                curBars = 0;
                countBars = 0;
                s.closedOrders.Clear();
                initialized = true;
                entrySignal = 0;
                if (!optimize)
                    LogPrint(" Init: Initialization successful. version=" + version);
            }
            catch (Exception e)
            {
                LogPrint("Init:" + e.Message);
                LogPrint(e.StackTrace);
            }
            return 0;
        }

        //+------------------------------------------------------------------+"
        //| expert deinitialization function                                 |"
        //|                                                                  |"
        //| called by Alveo to unitialize EA before termination.             |"
        //+------------------------------------------------------------------+"
        protected override int Deinit()
        {
            if (!optimize)
                LogPrint(strategy + " Deinit.");
            if (s != null)
                s = null;
            return 0;
        }

        // eternal interface for Simulator
        internal int doStart()
        {
            return this.Start();
        }

        //+------------------------------------------------------------------+"
        //| expert start function                                            |"
        //|                                                                  |"
        //| called by Alveo for every new bar on chart                       |"
        //| and maybe more often than that.                                  |"
        //+------------------------------------------------------------------+"
        protected override int Start()
        {
            try
            {
                var nBars = GetBars();      // When a new Bar closes
                if (nBars <= curBars)       // no new Bar
                    return 0;
                curBars = nBars;
                var oos = IsOutOfSession();
                if (!CheckParameters())  // make sure that all User Settings are valid
                {
                    if (!simulate)
                        Sleep(500);
                    if (!optimize)
                        LogPrint("User Settings are invalid !!");
                    return 0;
                }
                curTime = GetCurTime();
                curBar = GetCurBar();
                s.dI = curBar;
                curPrice = s.dI.close;
                midPrice = (s.dI.high + s.dI.low) / 2.0;
                symbol = GetSymbol();
                accountBalance = GetAccoutBalance();
                riskLimit = 1.8 * accountBalance / 100 / 10; // 1.8 percent risk limit in Pips for 1 Standard lot
                Monitor();
                if (DetectChanged(ref s.OKtoTrade, CheckOKToTrade()))
                {
                    if (!optimize)
                        LogPrint(" OKtoTrade=" + s.OKtoTrade.ToString());
                    startSession = true;
                }
                if (startSession)
                    countBars = 0;
                else
                    countBars++;
                if (!s.OKtoTrade)
                    return 0; // EA cannot trade at this time
                // else OKtoTrade
                Strategy();                 // execute trading startegy
                int orderID = 0;
                total = GetTotalOrders();
                if (!optimize)
                {
                    if (total > 0)
                    {
                        if (s.buyOpenOrders.Keys.Count > 0)
                            orderID = (int)s.buyOpenOrders.Keys.First();
                        else if (s.sellOpenOrders.Keys.Count > 0)
                            orderID = (int)s.sellOpenOrders.Keys.First();
                    }
                    int emaCrossed = (ema1.value > ema2.value && ema1.prevValue <= ema2.prevValue) ? 1
                        : (ema1.value < ema2.value && ema1.prevValue >= ema2.prevValue) ? -1 : 0;
                    string header = "dateTime, close, high, low, ema1, ema2, emaCrossed, rsi, crossFromAbove, crossFromBelow, entrySignal, numOrders, orderID";
                    string data = curTime
                        + ", " + curBar.close
                        + ", " + curBar.high
                        + ", " + curBar.low
                        + ", " + ema1.value
                        + ", " + ema2.value
                        + ", " + emaCrossed
                        + ", " + rsi.value
                        + ", " + rsi.crossFromAbove
                        + ", " + rsi.crossFromBelow
                        + ", " + entrySignal
                        + ", " + total
                        + ", " + ((orderID > 0) ? orderID.ToString() : "")
                        ;
                    WriteDebugLog(data, clear: initialized, header: header); // write traded data from order.Comment to TradeLog file
                }
                initialized = false;
                startSession = false;       // start of session is complete
            }
            catch (Exception e)  // in case something goes horribly wrong
            {
                LogPrint(" Start: Exception: " + e.Message);
                LogPrint("Exception: " + e.StackTrace);
                throw e;
            }
            return 0;
        }

        internal class EA_State
        {
            const string VERSION = "1.0";  //Restore file format
            static HLHB_Trend ea;
            internal bool firstRun;
            internal string symbol;
            internal bool isConnected;
            internal bool Stopped;
            internal bool isTradeAllowed;
            internal bool OKtoTrade;
            internal Statistics stats;
            internal BarData dI;
            internal int nBars = 0;
            internal DateTime OldTime, OldTime2, OldTime3, nextDay;
            internal int targetDir;
            internal double startingBalance;
            internal int ordersThisHour;
            private Object thisLock;
            internal Dictionary<long, Order> buyOpenOrders = null;
            internal Dictionary<long, Order> sellOpenOrders = null;
            internal Dictionary<long, Order> closedOrders = null;

            /// <summary>
            /// DENA3_State Class object constructor
            /// </summary>
            public EA_State()
            {
                firstRun = true;
                isConnected = true;
                Stopped = false;
                isTradeAllowed = false;
                OKtoTrade = true;
                stats = new Statistics(0);
                nBars = 0;
                thisLock = new Object();
                buyOpenOrders = null;
                sellOpenOrders = null;
                closedOrders = null;
            }

            internal EA_State(HLHB_Trend theEA) : this()
            {
                ea = theEA;
            }

            /// <summary>
            /// DENA3_State initialization
            /// </summary>
            internal void ClearState()
            {
                if (firstRun && !ea.optimize)
                {
                    lock (thisLock)
                    {
                        string path = ea.dataFileDir;
                        if (!System.IO.Directory.Exists(path))
                            System.IO.Directory.CreateDirectory(path);
                        var filename = path + "Journal.txt";
                        System.IO.File.Delete(filename);
                    }
                }
                nBars = 0;
                isConnected = true;
                Stopped = false;
                isTradeAllowed = false;
                stats = new Statistics(0);
                dI = null;
                targetDir = 0;
                startingBalance = double.MinValue;
                ordersThisHour = 0;
                OldTime = UtcTimeNow();
                OldTime2 = OldTime;
                OldTime3 = UtcTimeNow();
                nextDay = new DateTime(
                      OldTime3.Year,
                      OldTime3.Month,
                      OldTime3.Day,
                      0, 0, 0, 0) + TimeSpan.FromHours(24);
                buyOpenOrders = new Dictionary<long, Order>();
                sellOpenOrders = new Dictionary<long, Order>();
                closedOrders = new Dictionary<long, Order>();
            }

            /// <summary>
            /// Save DENA3_State to restore file
            /// <param name="restore">ref to System.IO.StreamWriter</param>
            /// </summary>
            public struct Statistics
            {
                internal double profitLoss;
                internal double maxProfit;
                internal double maxLoss;
                internal double maxDrawdown;
                internal double maxDailyDrawdown;
                internal double largestDailyDrawdown;
                internal bool exceededDailyDrawdown;
                internal int numContracts;
                internal int numWins;
                internal int numLoss;
                internal double avgWin;
                internal double avgLoss;
                internal double winRate;
                internal double expectancy;
                internal int numOpen;
                internal int numDays;
                internal int numHours;
                internal int numTrades;
                internal int numOrders;
                internal int orderExits;
                internal int computeStats;
                internal double sumPips;
                internal int sumWins;
                internal int sumLoss;
                internal double avgWin2;
                internal double avgLoss2;
                internal double avgWinBars;
                internal double avgLossBars;
                internal int numEntries;
                internal int numJISO;

                /// <summary>
                /// Initialize Trading Statistics
                /// </summary>
                internal Statistics(int x)
                {
                    profitLoss = 0;
                    maxProfit = 0;
                    maxLoss = 0;
                    maxDrawdown = 0;
                    maxDailyDrawdown = 0;
                    largestDailyDrawdown = 0;
                    exceededDailyDrawdown = false;
                    numContracts = 0;
                    numWins = 0;
                    numLoss = 0;
                    avgWin = 0;
                    avgLoss = 0;
                    winRate = 0;
                    expectancy = 0;
                    numOpen = 0;
                    numDays = 0;
                    numHours = 0;
                    numTrades = 0;
                    numOrders = 0;
                    orderExits = 0;
                    computeStats = 0;
                    sumPips = 0;
                    sumWins = 0;
                    sumLoss = 0;
                    avgWin2 = 0;
                    avgLoss2 = 0;
                    avgWinBars = 0;
                    avgLossBars = 0;
                    numEntries = 0;
                    numJISO = 0;
                }
            }
        }

        public class BarData
        {
            internal string symbol;
            internal Bar bar;
            internal DateTime BarTime;
            internal double open;
            internal double high;
            internal double low;
            internal double close;
            internal long volume;
            internal double bid;
            internal double ask;
            internal bool startOfSession;
            internal bool buySide;
            internal double typical;
            internal double change;
            internal double pctChange;
            internal double spread;
            internal DayOfWeek wd;
            internal TimeSpan tod;

            /// <summary>
            /// EA BarData Item Class
            /// </summary>
            public BarData()
            {
                BarData_Init();
            }

            /// <summary>
            /// BarData object constructor with Bar data
            /// </summary>
            public BarData(Bar ibar, double iBid = 0, double iAsk = 0)
            {
                BarData_Init();
                bar = ibar;
                BarTime = ibar.BarTime;
                open = (double)ibar.Open;
                high = (double)ibar.High;
                low = (double)ibar.Low;
                close = (double)ibar.Close;
                volume = ibar.Volume;
                wd = BarTime.DayOfWeek;
                tod = BarTime.TimeOfDay;
                bid = iBid;
                ask = iAsk;
                if (bid * ask > 0)  // if bid and ask are both > 0
                    spread = ask - bid;
            }

            /// <summary>
            /// BarData initialization
            /// </summary>
            internal void BarData_Init()
            {
                symbol = "";
                BarTime = UtcTimeNow();
                high = 0;
                low = 0;
                close = 0;
                volume = -1;
                bid = 0;
                ask = 0;
                startOfSession = true;
                buySide = true;
                typical = 0;
                change = 0;
                pctChange = 0;
                spread = 0;
                wd = BarTime.DayOfWeek;
                tod = BarTime.TimeOfDay;
            }
        }

        // Monitor for Closed orders
        // Monitor for Start of Day and start of Hour
        // Simulate market actions
        public void Monitor()
        {
            int total = GetTotalOrders();
            // count trading numDays
            if (s.OldTime3 > curTime || curTime >= s.nextDay)
            {
                if (!optimize)
                    LogPrint(" Start of Day. ");
                s.stats.numDays++;
                s.OldTime3 = curTime;
                s.nextDay = new DateTime(
                  s.OldTime3.Year,
                  s.OldTime3.Month,
                  s.OldTime3.Day,
                  0, 0, 0, 0) + TimeSpan.FromHours(24);
                var pctDd = 100 * s.stats.maxDailyDrawdown / accountBalance;  // update statistics
                if (s.stats.largestDailyDrawdown > pctDd)
                    s.stats.largestDailyDrawdown = pctDd;
                s.startingBalance = accountBalance;
                s.stats.maxDailyDrawdown = 0;
                s.stats.exceededDailyDrawdown = false;
                riskLimitReached = false;
            }
            // count trading numHours
            if (s.OldTime2 > curTime || curTime.Subtract(s.OldTime2) >= TimeSpan.FromHours(1))
            {
                if (!optimize)
                    LogPrint(" New Hour. ");
                s.stats.numHours++;
                s.OldTime2 = curTime;
                s.ordersThisHour = 0;
                if (!simulate)
                    LogPrint(" Monitor still running. " + curTime.ToLongTimeString());
            }
            var drawdown = accountBalance - s.startingBalance;
            if (drawdown < s.stats.maxDailyDrawdown)
            {
                s.stats.maxDailyDrawdown = drawdown;
            }
            var pctDrawdown = 100 * s.stats.maxDailyDrawdown / s.startingBalance;
            if (Math.Abs(pctDrawdown) > 4.9)
            {
                if (!optimize)
                    LogPrint(" CreateOrder: exceededDailyDrawdown " + pctDrawdown.ToString("F2"));
                s.stats.exceededDailyDrawdown = true;
            }
            if (s.stats.exceededDailyDrawdown)
            {
                CloseAllTrades(reason: 3);
                return; // wait until next day to trade
            }
            RemoveClosedOrders();
            var highest = curBar.high;
            var lowest = curBar.low;
            if (simulate)
            {
                // check buyOpenOrders for TP or SL
                if (s.buyOpenOrders.Count > 0)
                {
                    foreach (var order in s.buyOpenOrders.Values)
                    {
                        double tp = (double)order.TakeProfit;
                        double sl = (double)order.StopLoss;

                        TPflag = (tp > 0 && highest >= tp);
                        SLflag = (sl > 0 && lowest <= sl);
                        if (TPflag && SLflag)  // randomly reset one flag, if both are set
                        {
                            var nxtRnd = rnd.Next(1, 10);
                            var selectTP = (nxtRnd > 5);
                            if (selectTP)
                                SLflag = false;
                            else
                                TPflag = false;
                        }
                        if (TPflag)
                        {
                            if (!optimize)
                                LogPrint("**Buy Order TakeProfit. Ticket=" + order.Id);
                            ExitOpenTrade(reason: 1, order: order, price: tp + 1 * point);
                            if (simulate)
                                order.CloseType = CloseType.TakeProfit;
                        }
                        else if (SLflag)
                        {
                            if (!optimize)
                                LogPrint("!!Buy Order StopLoss. Ticket=" + order.Id);
                            ExitOpenTrade(reason: 2, order: order, price: sl - 1 * point);
                            if (simulate)
                                order.CloseType = CloseType.StopLoss;
                        }
                    }
                }

                // check sellOpenOrders for TP or SL
                if (s.sellOpenOrders.Count > 0)
                {
                    foreach (var order in s.sellOpenOrders.Values)
                    {
                        double tp = (double)order.TakeProfit;
                        double sl = (double)order.StopLoss;
                        TPflag = (tp > 0 && lowest <= tp);
                        SLflag = (sl > 0 && highest >= sl);
                        if (TPflag && SLflag)  // randomly reset one flag
                        {
                            var selectTP = (rnd.Next(1, 10) > 5);
                            if (selectTP)
                                SLflag = false;
                            else
                                TPflag = false;
                        }
                        if (TPflag)
                        {
                            if (!optimize)
                                LogPrint("**Sell Order TakeProfit. Ticket=" + order.Id);
                            ExitOpenTrade(reason: 3, order: order, price: tp - 1 * point);
                            if (simulate)
                                order.CloseType = CloseType.TakeProfit;
                        }
                        if (SLflag)
                        {
                            if (!optimize)
                                LogPrint("!!Sell Order StopLoss. Ticket=" + order.Id);
                            ExitOpenTrade(reason: 4, order: order, price: sl + 1 * point);
                            if (simulate)
                                order.CloseType = CloseType.StopLoss;
                        }
                    }
                }
            }
            RemoveClosedOrders();
            // check Pending orders
            List<long> deletedOrders = new List<long>();

            // check closed buyOpenOrders for Trailing StopLoss
            if (s.buyOpenOrders.Count > 0)
            {
                foreach (var ticketID in s.buyOpenOrders.Keys)
                {
                    Order order;
                    if (!s.buyOpenOrders.TryGetValue(ticketID, out order))
                        continue;
                    var tm = GetOrderCloseTime((int)ticketID);
                    if (tm.DateTime.Year > 1980) // Closed
                    {
                        if (simulate)
                            order.CloseType = CloseType.Market;
                        var dProfit = TradeClosed(order);
                        if (!optimize)
                            LogPrint("FT: Monitor: Closed Buy order. ID=" + order.Id
                                + " CloseTime=" + tm.ToString());
                    }
                    else // !closed
                    {
                        // update Trailing Stoploss
                        if (TrailingSL && (double)order.Quantity > 0)
                        {
                            CheckOrderModify((int)order.Id, StopLoss * 10, ref order);
                        }
                    }
                }
            }

            // check closed sellOpenOrders for Trailing StopLoss
            if (s.sellOpenOrders.Count > 0)
            {
                foreach (var ticketID in s.sellOpenOrders.Keys)
                {
                    Order order;
                    if (!s.sellOpenOrders.TryGetValue(ticketID, out order))
                        continue;
                    var tm = GetOrderCloseTime((int)ticketID);
                    if (tm.DateTime.Year > 1980) // Closed
                    {
                        var dProfit = TradeClosed(order);
                        if (!optimize)
                            LogPrint("Monitor: Closed Sell order. ID=" + order.Id
                            + " CloseTime=" + tm.ToString());
                    }
                    else // !closed, update trailing Stoploss
                    {
                        if (TrailingSL && (double)order.Quantity > 0)
                        {
                            CheckOrderModify((int)order.Id, StopLoss * 10, ref order);
                        }
                    }
                }
            }
            RemoveClosedOrders();
        }

        internal void Strategy()
        {
            try
            {
                if (startSession)                                               // on startSession, Close all Open trades
                {
                    if (!optimize)
                        LogPrint(" startSession=" + startSession);
                    CloseAllTrades(reason: 1);
                }
                entrySignal = GetEnterSignal();                                 // get entrySignal, 0=no signal, 1=Buy signal, -1=Sell signal
                if (entrySignal != 0 && entrySignal != s.targetDir)             // if entrySignal has changed direction
                {
                    total = GetTotalOrders();                                   // get number of Open trades
                    if (total > 0)                                              // Close any Open trades
                        CloseAllTrades(reason: 2, clear: true, sidemask: 3);   
                    paused = CheckPaused();                                     // Do not Enter new trades while trading is Paused
                    total = GetTotalOrders();                                   
                    if (!paused && total == 0)                                  // if no trades Open and !paused
                    {
                        s.targetDir = entrySignal;                              // Enter Market trade in the direction of the entrySignal
                        var ticket = CreateOrder(lotsize: Lotsize, stoploss: StopLoss * 10, takeprofit: TakeProfit * 10);
                    }
                }
            }
            catch (Exception e)
            {
                LogPrint(" Exception: " + e.Message);
                LogPrint(" Exception: " + e.StackTrace);
            }
        }

        internal int GetEnterSignal()               // GetEnterSignal will return int entrySignal direction, 0=no signal, 1=Buy signal, -1=Sell signal
        {
            int signal = 0;                         // no signal
            rsi.Calc(midPrice);                     // update indicator values
            ema1.Calc(curPrice);
            ema2.Calc(curPrice);
            if (countBars > 5)                      // if at least 5 bars after startSession
            {
                if (ema1.value > ema2.value         // EMA1 crossed above EMA2
                    && ema1.prevValue <= ema2.prevValue
                    && rsi.crossFromBelow)          // and RSI has crossed 50 from Below within last 2 Bars
                    signal = 1;                     // signal Side = Buy
                else if (ema1.value < ema2.value    // EMA1 crossed below EMA2
                    && ema1.prevValue >= ema2.prevValue
                    && rsi.crossFromAbove)          // and RSI has crossed 50 from Above within last 2 Bars
                    signal = -1;                    // signal Side = Sell
            }
            return signal;
        }

        internal void WriteData(string msg, string filename)
        {
            using (System.IO.StreamWriter file =
                       new System.IO.StreamWriter(dataFileDir + filename, append: true))
            {
                file.WriteLine(msg);
                file.Close();
            }
        }

        internal bool CheckPaused()
        {
            var LT = GetLocalCurTime();
            var prevPaused = paused;
            paused = ((LT.DayOfWeek == System.DayOfWeek.Friday && LT.TimeOfDay >= fridayPause)
                        || (LT.TimeOfDay > dailyMaintStart && LT.TimeOfDay < dailyMaintEnd)
                );
            if (paused)
            {
                if (!optimize && !prevPaused)
                    LogPrint(" Trading Paused. Friday afternoon or Alveo maintenance, local time=" + LT.ToString());
                if (LT.TimeOfDay > dailyMaintEnd)
                {
                    paused = false;
                }
            }
            if (!paused)
            {
                if (!optimize && prevPaused)
                    LogPrint(" Trading Resumed. local time=" + LT.ToString());
            }
            return paused;
        }

        internal void CheckOrderModify(int ticket, int SLlimit, ref Order order)
        {
            ModifyOrder(ticket, SLlimit, ref order);
        }

        internal void RemoveClosedOrders()  // Remove Closed ordrs from Open Orders Lists
        {
            if (s.closedOrders.Count > 0)
            {
                foreach (var iD in s.closedOrders.Keys)
                {
                    if (s.buyOpenOrders.Count > 0)
                        s.buyOpenOrders.Remove(iD);
                    if (s.sellOpenOrders.Count > 0)
                        s.sellOpenOrders.Remove(iD);
                }
                s.closedOrders.Clear();
            }
        }

        internal bool CheckParameters()
        {
            if (optimize)
                return true;
            return true;
        }

        // external interface for Simulation Statistics
        internal EA_State.Statistics GetStats()
        {
            EA_State.Statistics stats = s.stats;
            return stats;
        }

        internal struct MqlDateTime
        {
            internal int year;           // Year
            internal int mon;            // Month
            internal int day;            // Day
            internal int hour;           // Hour
            internal int min;            // Minutes
            internal int sec;            // Seconds
            internal int day_of_week;    // Day of week (0-Sunday, 1-Monday, ... ,6-Saturday)
            internal int day_of_year;    // Day number of the year (January 1st is assigned the number value of zero)
        };

        internal void TimeToStruct(
            DateTime dt,            // date and time
            ref MqlDateTime dt_struct      // structure for the adoption of values
        )
        {
            dt_struct.year = dt.Year;           // Year
            dt_struct.mon = dt.Month;            // Month
            dt_struct.day = dt.Day;            // Day
            dt_struct.hour = dt.Hour;           // Hour
            dt_struct.min = dt.Minute;            // Minutes
            dt_struct.sec = dt.Second;            // Seconds
            dt_struct.day_of_week = (int)dt.DayOfWeek;    // Day of week (0-Sunday, 1-Monday, ... ,6-Saturday)
            dt_struct.day_of_year = dt.DayOfYear;    // Day number of the year (January 1st is assigned the number value of zero)
        }

        int PeriodSeconds(TimeFrame period = TimeFrame.Unknown)     // chart period
        {
            int seconds = (period != TimeFrame.Unknown) ? (int)period * 60 : 1;
            return seconds;
        }

        bool IsOutOfSession()
        {
            MqlDateTime time0 = new MqlDateTime();
            var curTime = (datetime)GetCurTime();
            TimeToStruct(curTime, ref time0);
            int weekDay = time0.day_of_week;
            long timeFromMidnight = curTime % 86400;
            int periodLength = PeriodSeconds(GetTimeframe());
            bool skipTrade = false;

            if (weekDay == 0)
            {
                if (sessionIgnoreSunday) return true;
                int lastBarFix = sessionCloseAtSessionClose ? periodLength : 0;
                skipTrade = timeFromMidnight < sessionSundayOpen || timeFromMidnight + lastBarFix > sessionSundayClose;
            }
            else if (weekDay < 5)
            {
                int lastBarFix = sessionCloseAtSessionClose ? periodLength : 0;
                skipTrade = timeFromMidnight < sessionMondayThursdayOpen || timeFromMidnight + lastBarFix > sessionMondayThursdayClose;
            }
            else
            {
                int lastBarFix = sessionCloseAtFridayClose || sessionCloseAtSessionClose ? periodLength : 0;
                skipTrade = timeFromMidnight < sessionFridayOpen || timeFromMidnight + lastBarFix > sessionFridayClose;
            }
            return (skipTrade);
        }


        // for optimizing, load trading parameters from file
        void LoadBestParameters()
        {
            try
            {
                string path;
                string parseErr = "Exception LoadBestParameters: parse error at ";
                string paramFilename = "ParamSet" + symbol.Replace("/", "") + ((simulate) ? "0" : "1") + ".csv";
                path = dataFileDir + paramFilename;
                if (!System.IO.File.Exists(path))
                {
                    if (!optimize)
                        LogPrint("EA: File " + path + " does not exist !!");
                    throw new Exception("Exception: File " + path + " does not exist !!");
                }
                System.IO.StreamReader best = new System.IO.StreamReader(path);
                if (!optimize)
                    LogPrint("EA: Reading File " + path);
                var line = best.ReadLine();
                line = best.ReadLine();
                char[] delimiterChars = { ',', '\t' };
                string[] vals = line.Split(delimiterChars);
                if (vals.Count() != 19)
                    throw new Exception("Exception: File " + path + " improper format !!");
                // var digits = GetDigits();
                strategy = vals[0];
                int X;
                if (!int.TryParse(vals[4], out X))
                    throw new Exception(parseErr + "EMA1_period !!");
                EMA1_period = X;
                if (!int.TryParse(vals[5], out X))
                    throw new Exception(parseErr + "EMA2_period !!");
                EMA2_period = X;
                if (!int.TryParse(vals[6], out X))
                    throw new Exception(parseErr + "StopLoss !!");
                StopLoss = X;
                if (!int.TryParse(vals[7], out X))
                    throw new Exception(parseErr + "TakeProfit !!");
                TakeProfit = X;
                if (!int.TryParse(vals[8], out X))
                    throw new Exception(parseErr + "RSI_period !!");
                RSI_period = X;
            }
            catch (Exception e)
            {
                LogPrint("Exception: " + e.Message);
                LogPrint("Exception: " + e.StackTrace);
            }
            return;
        }

        void JournalOnExit(int reason, Order order, double bclose, BarData dI,
        double t2 = 0, double t5 = 0, double dPofit = 0, double stoploss = 0)
        {
            if (optimize || simulate || order == null || true)
                return;
            var side = order.Side;

            double openPrice = (double)order.OpenPrice;
            switch (reason)
            {
                case 1: // take profit on Buy order
                    LogPrint("ExitOpenTrade: reason = take profit on Buy order.");
                    LogPrint(" order=" + order.Id + " openPrice=" + openPrice.ToString("F5")
                        + " close=" + bclose.ToString("F5") + " dPofit=" + dPofit.ToString("F5")
                    );
                    break;
                case 2: // stoploss on Buy order
                    LogPrint("ExitOpenTrade: reason = stoploss on Buy order.");
                    LogPrint(" order=" + order.Id + " openPrice=" + openPrice.ToString("F5")
                        + " close=" + bclose.ToString("F5") + " dPofit=" + dPofit.ToString("F5"));
                    break;
                case 3: // take profit on Sell order
                    LogPrint("ExitOpenTrade: reason = take profit on Sell order.");
                    LogPrint(" order=" + order.Id + " openPrice=" + openPrice.ToString("F5")
                        + " close=" + bclose.ToString("F5") + " dPofit=" + dPofit.ToString("F5")
                    );
                    break;
                case 4: // stoploss on Sell order
                    LogPrint("ExitOpenTrade: reason = stoploss on Sell order.");
                    LogPrint(" order=" + order.Id + " openPrice=" + openPrice.ToString("F5")
                        + " close=" + bclose.ToString("F5") + " dPofit=" + dPofit.ToString("F5"));
                    break;
                case 5: // closeAllTrades: tradingclosed || riskLimitReached
                    LogPrint("closeAllTrades: reason = tradingclosed || riskLimitReached close=" + bclose.ToString());
                    LogPrint(" dI.Date=" + dI.BarTime.ToShortDateString() + " " + dI.BarTime.ToLongTimeString() + " dPofit=" + dPofit.ToString("F5"));
                    LogPrint(" tradingclosed=" + tradingclosed + " riskLimitReached=" + riskLimitReached);
                    break;
                case 6: // closeAllTrades: timeToExit
                    LogPrint("closeOpenTrade: reason = timeToExit  close=" + bclose.ToString());
                    LogPrint(" dI.Date=" + dI.BarTime.ToShortDateString() + " " + dI.BarTime.ToLongTimeString() + " dPofit=" + dPofit.ToString("F5"));
                    break;
                case 7: // closeAllTrades: !isTradeAllowed
                    LogPrint("closeOpenTrade: reason = !isTradeAllowed. close=" + bclose.ToString() + " isTradeAllowed=" + s.isTradeAllowed.ToString());
                    LogPrint(" dI.Date=" + dI.BarTime.ToShortDateString() + " " + dI.BarTime.ToLongTimeString() + " dPofit=" + dPofit.ToString("F5"));
                    break;
                case 8: // closeAllTrades: IsEaStopped
                    LogPrint("closeOpenTrade: reason = IsEaStopped.  bclose=" + bclose.ToString() + " IsEaStopped=" + IsEaStopped.ToString());
                    LogPrint(" dI.Date=" + dI.BarTime.ToShortDateString() + " " + dI.BarTime.ToLongTimeString() + " dPofit=" + dPofit.ToString("F5"));
                    break;
                case 9: // IsEaStopped
                    LogPrint("closeOpenTrade: reason = IsEaStopped.  bclose=" + bclose.ToString() + " IsEaStopped=" + IsEaStopped.ToString());
                    LogPrint(" dI.Date=" + dI.BarTime.ToShortDateString() + " " + dI.BarTime.ToLongTimeString() + " dPofit=" + dPofit.ToString("F5"));
                    break;
                case 99:  // closeAllTrades at end of simulation
                    LogPrint("closeOpenTrade: reason = closeAllTrades at end of simulation. close=" + bclose.ToString());
                    LogPrint(" dI.Date=" + dI.BarTime.ToShortDateString() + " " + dI.BarTime.ToLongTimeString() + " dPofit=" + dPofit.ToString("F5"));
                    break;
                default:
                    throw new Exception("Exception: JournalOnclose reason not defined !! ");
                    //break;
            }
        }

        internal void ComputeStats(double dPofit, double dPips, double dBars, Order order)
        {
            s.stats.sumPips += dPips;
            s.stats.profitLoss += dPofit;  // update profitLoss
            simAccountBalance += dPofit;
            if ((double)order.Quantity > 0)
            {
                s.stats.numContracts++;  // update numContracts
                if (dPofit >= 0)
                {
                    double count = (double)s.stats.sumWins;
                    s.stats.sumWins++;
                    s.stats.avgWin2 = (s.stats.avgWin2 * count + dPofit) / (double)s.stats.sumWins;
                    s.stats.avgWinBars = (s.stats.avgWinBars * count + dBars) / (double)s.stats.sumWins;
                }
                else // dPofit < 0
                {
                    double count = (double)s.stats.sumLoss;
                    s.stats.sumLoss++;
                    s.stats.avgLoss2 = (s.stats.avgLoss2 * count + dPofit) / (double)s.stats.sumLoss;
                    s.stats.avgLossBars = (s.stats.avgLossBars * count + dBars) / (double)s.stats.sumLoss;
                }

                // calculate avgWin, avgLoss, numWins, numLoss, maxProfit, maxLoss, maxDrawdown
                if (dPofit >= 0)
                {
                    s.stats.avgWin = (s.stats.numWins * s.stats.avgWin + Math.Abs(dPofit)) / (s.stats.numWins + 1);
                    s.stats.numWins++;
                }
                else // (dPofit <= 0), Loss
                {
                    s.stats.avgLoss = (s.stats.numLoss * s.stats.avgLoss + Math.Abs(dPofit)) / (s.stats.numLoss + 1);
                    s.stats.numLoss++;
                }
            }
            if (dPofit > s.stats.maxProfit)
                s.stats.maxProfit = dPofit;
            else if (dPofit < s.stats.maxLoss)
                s.stats.maxLoss = dPofit;
            if (s.stats.profitLoss < s.stats.maxDrawdown) // compute maxDrawdown
                s.stats.maxDrawdown = s.stats.profitLoss;
            s.stats.computeStats++; // count num Order statistics computed
        }

        internal void WriteTradeLog(string msg, bool clear = false, bool countLine = false, string header = "")
        {
            lock (TLock)  // exclussive lock, one call access at a time
            {
                // Write msg to TradeLog, Delete and Create new file is clear=true
                var tf = GetTimeframe();
                var tfStr = (tf != TimeFrame.Unknown) ? tf.ToString() : "S1";
                var account = GetAccountNumber();
                string tradeLogFilename = strategy + " TradeLog_" + symbol.Replace("/", "") + "_" + tfStr + "_" + account + ".csv";
                if (!System.IO.Directory.Exists(dataFileDir))
                    System.IO.Directory.CreateDirectory(dataFileDir);
                if (clear)
                    System.IO.File.Delete(dataFileDir + tradeLogFilename);
                tradeLogFile = new System.IO.StreamWriter(dataFileDir + tradeLogFilename, append: true);
                if (clear && header.Length > 0)
                    tradeLogFile.WriteLine(header);
                tradeLogFile.WriteLine(msg);
                tradeLogFile.Close();
            }
        }

        internal void WriteDebugLog(string msg, bool clear = false, string header = "")
        {
            lock (TLock)  // exclussive lock, one call access at a time
            {
                // Write msg to TradeLog, Delete and Create new file is clear=true
                var tf = GetTimeframe();
                var tfStr = (tf != TimeFrame.Unknown) ? tf.ToString() : "S1";
                var account = GetAccountNumber();
                string debugLogFilename = strategy + " DebugLog" + symbol.Replace("/", "") + "_" + tfStr + ".csv";
                if (!System.IO.Directory.Exists(dataFileDir))
                    System.IO.Directory.CreateDirectory(dataFileDir);
                if (clear)
                    System.IO.File.Delete(dataFileDir + debugLogFilename);
                debugLogFile = new System.IO.StreamWriter(dataFileDir + debugLogFilename, append: true);
                if (clear && header.Length > 0)
                    debugLogFile.WriteLine(header);
                debugLogFile.WriteLine(msg);
                debugLogFile.Close();
            }
        }

        #region  Alveo simulation routines
        // emulate Alveo Bars
        internal int GetBars()
        {
            if (simulate)
                return simBars;
            else // Alveo
                return Bars;
        }

        // Get most recent Bar time
        public DateTime GetLocalCurTime()
        {
            if (simulate)
                return simLocalTime;
            else
                return DateTime.Now;
        }

        public DateTime GetCurTime()
        {
            if (simulate)
                return simTime;
            else
                return UtcTimeNow();
        }

        // interface to get current UTC time just in case Alveo has a problem with the function
        static DateTime UtcTimeNow()
        {
            DateTime utcNow = DateTime.UtcNow;
            return utcNow;
        }

        // Get Chart currency Symbol
        internal string GetSymbol()
        {
            string symbol = simSymbol;
            if (!simulate)
            {
                symbol = Symbol();
            }
            return symbol;
        }

        // Get Account Balance
        public double GetAccoutBalance()
        {
            if (!simulate)
            {
                double accountBalance = AccountBalance();
                return accountBalance;
            }
            return simAccountBalance;  //simulated AccountBalance
        }

        /// <summary>
        /// LogPrint: print msg to Alveo Log or to Console
        /// </summary>
        /// <returns>void</returns>
        internal void LogPrint(string msg, bool doIt = false)
        {
            if (optimize && !doIt)
                return;
            var tf = GetTimeframe();
            var tfStr = (tf != TimeFrame.Unknown) ? tf.ToString() : "S1";
            string msg2 = strategy + " [" + symbol + " " + tfStr + "]: " + msg;
            if (!simulate)
                Print(msg2);
            else
                Console.WriteLine(msg2);
        }

        // if value != flag, then change flag = value and return true, otherwise return flase (not changed)
        bool DetectChanged(ref bool flag, bool value)
        {
            bool changed;
            changed = (value != flag);
            if (changed)
            {
                flag = value;
            }
            return changed;
        }

        // checked conditions to determine if it is OK to Trade
        internal bool CheckOKToTrade()
        {
            if (accountBalance > 0 && accountBalance < 50.00)
            {
                if (!optimize)
                    LogPrint(" balance is too low !! AccountBalance = " + accountBalance.ToString("F2"));
                if (!simulate)
                    Sleep(1000);
                return false;
            }

            if (DetectChanged(ref riskLimitReached, (s.startingBalance >= 0) ? accountBalance / s.startingBalance < 0.951 : false))
            {
                if (riskLimitReached)
                {
                    if (!optimize)
                    {
                        string msg = (" Start: riskLimitReached, trading prevented. "
                        + " startingBalance =" + s.startingBalance.ToString("F2")
                        + " accountBalance =" + accountBalance.ToString("F2"));
                        LogPrint(msg);
                        LogPrint(msg);
                    }
                }
            }
            if (riskLimitReached)
            {
                CloseAllTrades(reason: 5);
                if (!simulate)
                    Sleep(1000); // sleep 1 sec
                return false;
            }


            if (DetectChanged(ref tradingclosed, CheckTradingClosed()))
            {
                if (tradingclosed)
                {
                    if (!optimize)
                    {
                        string msg = (" Start: tradingclosed. ");
                        LogPrint(msg);
                        LogPrint(msg);
                    }
                    CloseAllTrades(reason: 6);
                    if (!simulate)
                        Sleep(1000); // sleep 1 sec
                    return false;
                }
            }
            if (tradingclosed)
                return false;

            if (!simulate)  // Alveo functions
            {
                if (DetectChanged(ref s.isConnected, IsEAConnected()))
                {
                    if (!s.isConnected)
                    {
                        LogAlert("ALERT " + strategy + " EA [" + symbol + "]: not connected!");
                        if (!optimize)
                        {
                            string msg = ("not connected! ");
                            LogPrint(msg);
                        }
                        if (!simulate)
                            Sleep(1000); // sleep 1 sec
                        return false;  // not connected to Server
                    }
                    else // isConnected
                    {
                        if (!optimize)
                            LogPrint(" is now connected!");
                    }
                }
                if (!s.isConnected)
                    return false;

                if (DetectChanged(ref s.isTradeAllowed, IsTradeAllowed()))
                {
                    if (!s.isTradeAllowed)
                    {
                        if (!optimize)
                            LogPrint(" EA not allowed to Trade !");
                        //closeAllTrades(reason: 7);
                        if (!simulate)
                            Sleep(1000); // sleep 1 sec
                        return false; // EA not allowed to Trade
                    }
                    else // isTradeAllowed
                    {
                        if (!optimize)
                            LogPrint(" is now IsTradeAllowed!");
                    }
                }
                if (!s.isTradeAllowed)
                    return false;

                if (DetectChanged(ref s.Stopped, IsEaStopped))
                {
                    if (s.Stopped)
                    {
                        if (!optimize)
                            LogPrint(" EA stopped !");
                        //closeAllTrades(reason: 8);
                        if (!simulate)
                            Sleep(1000); // sleep 1 sec
                        return false; // EA Stopped
                    }
                    else
                    {
                        if (!optimize)
                            LogPrint(" EA resumed !");
                    }
                }

                if (IsEaStopped)
                {
                    //closeAllTrades(reason: 9);
                    if (!simulate)
                        Sleep(1000); // sleep 1 sec
                    return false;
                }
            }
            return true; //OKToTrade
        }

        // Get latest Chart Bar
        internal BarData GetCurBar()
        {
            BarData dI = null;
            if (simulate)
            {
                dI = curBar;  // data from Simulator
                var b = new Bar();
                b.BarTime = dI.BarTime;
                b.Open = (decimal)dI.open;
                b.High = (decimal)dI.high;
                b.Low = (decimal)dI.low;
                b.Close = (decimal)dI.close;
                b.Volume = dI.volume;
                dI.bar = b;
                dI.ask = dI.close + 5 * point;
                dI.bid = dI.close - 5 * point;
                dI.spread = dI.ask - dI.bid;
            }
            else
            {
                int last = (Bars < 2) ? 0 : 1;
                Bar b = ChartBars[last]; // last closed bar
                dI = new BarData(b, Bid, Ask);
                dI.bar = b;
            }
            return dI;
        }

        // Exit the specified Order at the specified price
        internal int ExitOpenTrade(int reason, Order order, double price = 0)
        {
            if (order == null)
                return -1;
            var orderID = (int)order.Id;
            var clsTime = GetOrderCloseTime(orderID);
            if (clsTime.DateTime.Year > 1970)
                return -1;  // already closed
            if (simulate && order.CloseDate == DateTime.MinValue)  // CloseDate not yet set
            {
                order.ClosePrice = (decimal)((price > 0) ? price : curPrice);
                order.CloseDate = curTime;
                order.CloseType = (reason == 50) ? CloseType.Market : CloseType.Unknown;
            }
            else if (!simulate)
            {
                var lots = (double)order.Quantity;
                if (!optimize)
                    LogPrint("Closd order=" + orderID.ToString() + " lots=" + lots.ToString());
                if (OrderSelect(orderID, SELECT_BY_TICKET) == true)
                {
                    if (!optimize)
                        LogPrint("Open price for order #" + orderID.ToString() + " is " + OrderOpenPrice().ToString());
                }
                else // OrderSelect returned err
                {
                    var err = GetLastError();
                    if (!optimize)
                        LogPrint("ExitOpenTrade: ticket=" + orderID + "  !!! OrderSelect returned error of " + ErrorDescription(err));
                    return -1;
                }
                // !PendingNew then Open Order, use OrderClose
                {
                    OrderClose(orderID, lots, 0, 0);
                    if (!simulate)
                        Sleep(500);
                }
            }
            if (order.CloseDate == DateTime.MinValue)  // CloseDate not yet set
            {
                order.CloseDate = GetOrderCloseTime(orderID);
            }
            s.stats.orderExits++; // count orderExits
            var dProfit = TradeClosed(order); // compute Order statistics
            if (!simulate)
            {
                LogPrint("closeOpenTrade orderID=" + orderID.ToString() + "  reason=" + reason.ToString()
                    + " price=" + order.ClosePrice.ToString("F5")
                    + " stoploss=" + order.StopLoss.ToString("F5")
                    + "\n");
                LogPrint(" profitLoss=" + s.stats.profitLoss.ToString("F5")
                    + " numWins=" + s.stats.numWins.ToString()
                    + " numLoss=" + s.stats.numLoss.ToString()
                    + " maxDrawdown=" + s.stats.maxDrawdown.ToString("F5")
                    );
                JournalOnExit(reason, order, bclose: (double)s.dI.close, dI: s.dI, dPofit: dProfit);
            }
            if (!optimize)
            {
                string msg = ("ExitOpenTrade: order=" + orderID + " reason=" + reason
                    + " dProfit=" + dProfit.ToString("F5")
                    + " balance=" + ((!simulate) ? accountBalance.ToString("F2") : simAccountBalance.ToString("F2")));
                LogPrint(msg);
            }

            return orderID;
        }

        // Specified Order was Closed, compute statistics
        public double TradeClosed(Order order)
        {
            if (order == null)
                return 0;
            if (order.Type == TradeType.Stop)
                return 0;
            var orderID = (int)order.Id;
            double dProfit = GetProfit(order);
            order.Profit = (decimal)dProfit;
            double dPips = GetPriceDiff(order) * pipPos;
            if (!simulate)
            {
                order.CloseDate = GetOrderCloseTime(orderID);
                order.ClosePrice = (decimal)GetOrderClosePrice(orderID, order);
            }
            var diffTime = order.CloseDate.Subtract(order.OpenDate);
            var diffMinutes = diffTime.TotalMinutes;
            double period = (double)PeriodSeconds(GetTimeframe()) / 60;
            double dBars = Math.Round(diffMinutes / period);
            if (!s.closedOrders.Keys.Contains(orderID))  // if orderID is not in closedOrders list
            {
                s.closedOrders.Add(orderID, order); // add Order to closedOrders list
                ComputeStats(dProfit, dPips, dBars, order); // compute statiustics
                if (!optimize)
                {
                    order.Comment +=  // add closing data to order.Comment
                        "," + orderID
                        + "," + order.Quantity
                        + "," + ((order.Side == TradeSide.Buy) ? "Buy" : "Sell")
                        + "," + order.OpenDate.ToString()
                        + "," + order.CloseDate.ToString()
                        + "," + order.OpenPrice.ToString("F5")
                        + "," + order.ClosePrice.ToString("F5")
                        + "," + order.StopLoss.ToString("F5")
                        + "," + order.TakeProfit.ToString("F5")
                        + "," + dProfit.ToString("F5");
                    var header = "comment, ID, Qty, Side, OpenDate,CloseDate, OpenPrice, ClosePrice, StopLoss, TakeProfit, dProfit, balance";
                    LogPrint("TradeClosed: orderID=" + orderID
                        + " side=" + order.Side
                        + " price=" + order.ClosePrice.ToString("F5")
                        + " dPofit=" + dProfit.ToString("F5"));
                    WriteTradeLog(order.Comment + "," + simAccountBalance.ToString("F2"), countLine: true, clear: clearFile, header: header); // write traded data from order.Comment to TradeLog file
                    clearFile = false;
                }
            }
            return dProfit;
        }

        /// <summary>
        /// IsEAConnected: calls Alveo IsConnected()
        /// </summary>
        /// <returns>bool IsConnected</returns>
        bool IsEAConnected()
        {
            if (!simulate)
                return IsConnected();
            else
                return true;
        }

        // Generate Alveo Alert or write msg to PC Console
        void LogAlert(string msg)
        {
            if (!simulate)
                Alert(msg);
            else
                Console.WriteLine(msg);
        }

        // Close All trades for the specified reason and optionally clear buyOpenOrders and sellOpenOrders lists
        internal void CloseAllTrades(int reason, bool clear = true, int sidemask = 0xF)
        {
            if (!optimize)
            {
                string msg = ("closeAllTrades: reason=" + reason + " curTime=" + curTime.ToShortDateString() + " " + curTime.ToLongTimeString());
                LogPrint(msg);
            }
            if (!simulate) // Alveo functions
            {
                int total = GetTotalOrders();
                Order order = null;
                for (int pos = 0; pos < total; pos++)
                {
                    if (OrderSelect(pos, SELECT_BY_POS) == false)
                        continue;
                    var cm = OrderComment();
                    if (!cm.Contains(magicStr))
                        continue;
                    var ticket = OrderTicket();  // get ticket (order.Id)
                    if (((sidemask & 0x1) > 0) && s.buyOpenOrders.ContainsKey(ticket))  // get Order data from buyOpenOrders list
                    {
                        order = s.buyOpenOrders[ticket];
                    }
                    else if (((sidemask & 0x2) > 0) && s.sellOpenOrders.ContainsKey(ticket)) // get Order data from sellOpenOrders list
                    {
                        order = s.sellOpenOrders[ticket];
                    }
                    else
                        continue;
                    ExitOpenTrade(reason, order, curPrice);  // Exit the specified Order for the specified reason at the curPrice
                }
            }
            else // simulate functions
            {
                if (((sidemask & 0x1) > 0) && s.buyOpenOrders.Count > 0)  // buyOpenOrders
                    foreach (var order in s.buyOpenOrders.Values)
                    {
                        ExitOpenTrade(reason, order, curPrice); // Exit the specified Order for the specified reason at the curPrice
                    }
                if (((sidemask & 0x2) > 0) && s.sellOpenOrders.Count > 0)  // buyOpenOrders
                    foreach (var order in s.sellOpenOrders.Values)
                    {
                        ExitOpenTrade(reason, order, curPrice); // Exit the specified Order for the specified reason at the curPrice
                    }
            }
            if (clear)
            {
                if ((sidemask & 0x1) > 0)
                    s.buyOpenOrders.Clear();
                if ((sidemask & 0x2) > 0)
                    s.sellOpenOrders.Clear();
            }
            if (!simulate)
                Sleep(1000); // sleep 1 sec
            return;
        }

        // if Time to Exit, then return true; otherwise false
        internal bool CheckTradingClosed()
        {
            bool closed = false;
            var tm = GetCurTime();
            var lt = GetLocalCurTime();
            closed = ((sessionIgnoreSunday && lt.DayOfWeek == System.DayOfWeek.Sunday)
                || tm.DayOfWeek == System.DayOfWeek.Saturday
                || (lt.DayOfWeek == System.DayOfWeek.Friday && lt.TimeOfDay > fridayclose)
                || (tm.DayOfWeek == System.DayOfWeek.Sunday && tm.TimeOfDay < sundayOpen)
                );
            return closed;
        }

        // convert Alveo int err to string errorDescription
        string ErrorDescription(int err)
        {
            string errorDescription = "err# " + err.ToString();
            switch (err)
            {
                case 0:
                    errorDescription += " ERR_NO_ERROR";
                    break;
                case 1:
                    errorDescription += " ERR_NO_RESULT";
                    break;
                case 2:
                    errorDescription += " ERR_COMMON_ERROR";
                    break;
                case 3:
                    errorDescription += " ERR_INVALID_TRADE_PARAMETERS";
                    break;
                case 4:
                    errorDescription += " ERR_SERVER_BUSY";
                    break;
                case 5:
                    errorDescription += " ERR_OLD_VERSION";
                    break;
                case 6:
                    errorDescription += " ERR_NO_CONNECTION";
                    break;
                case 7:
                    errorDescription += " ERR_NOT_ENOUGH_RIGHTS";
                    break;
                case 8:
                    errorDescription += " ERR_TOO_FREQUENT_REQUESTS";
                    break;
                case 9:
                    errorDescription += " ERR_MALFUNCTIONAL_TRADE";
                    break;
                case 64:
                    errorDescription += " ERR_ACCOUNT_DISABLED";
                    break;
                case 65:
                    errorDescription += " ERR_INVALID_ACCOUNT";
                    break;
                case 128:
                    errorDescription += " ERR_TRADE_TIMEOUT";
                    break;
                case 129:
                    errorDescription += " ERR_INVALID_PRICE";
                    break;
                case 130:
                    errorDescription += " ERR_INVALID_STOPS";
                    break;
                case 131:
                    errorDescription += " ERR_INVALID_TRADE_volume";
                    break;
                case 132:
                    errorDescription += " ERR_MARKET_closed";
                    break;
                case 133:
                    errorDescription += " ERR_TRADE_DISABLED";
                    break;
                case 134:
                    errorDescription += " ERR_NOT_ENOUGH_MONEY";
                    break;
                case 135:
                    errorDescription += " ERR_PRICE_CHANGED";
                    break;
                case 136:
                    errorDescription += " ERR_OFF_QUOTES";
                    break;
                case 137:
                    errorDescription += " ERR_BROKER_BUSY";
                    break;
                case 138:
                    errorDescription += " ERR_REQUOTE";
                    break;
                case 139:
                    errorDescription += " ERR_ORDER_LOCKED";
                    break;
                case 140:
                    errorDescription += " ERR_LONG_POSITIONS_ONLY_ALlowED";
                    break;
                case 141:
                    errorDescription += " ERR_TOO_MANY_REQUESTS";
                    break;
                case 145:
                    errorDescription += " ERR_TRADE_MODIFY_DENIED";
                    break;
                case 146:
                    errorDescription += " ERR_TRADE_CONTEXT_BUSY";
                    break;
                case 147:
                    errorDescription += " ERR_TRADE_EXPIRATION_DENIED";
                    break;
                case 148:
                    errorDescription += " ERR_TRADE_TOO_MANY_ORDERS";
                    break;
                case 149:
                    errorDescription += " ERR_TRADE_HEDGE_PROHIBITED";
                    break;
                case 150:
                    errorDescription += " ERR_TRADE_PROHIBITED_BY_FIFO";
                    break;
                case 4000:
                    errorDescription += " ERR_NO_MQLERROR";
                    break;
                case 4001:
                    errorDescription += " ERR_WRONG_FUNCTION_POINTER";
                    break;
                case 4002:
                    errorDescription += " ERR_ARRAY_INDEX_OUT_OF_RANGE";
                    break;
                case 4003:
                    errorDescription += " ERR_NO_MEMORY_FOR_CALL_STACK";
                    break;
                case 4013:
                    errorDescription += " ERR_ZERO_DIVIDE";
                    break;
                case 4014:
                    errorDescription += " ERR_UNKNOWN_COMMAND";
                    break;
                case 4015:
                    errorDescription += " ERR_WRONG_JUMP";
                    break;
                case 4016:
                    errorDescription += " ERR_NOT_INITIALIZED_ARRAY";
                    break;
                case 4017:
                    errorDescription += " ERR_DLL_CALLS_NOT_ALlowED";
                    break;
                case 4018:
                    errorDescription += " ERR_CANNOT_LOAD_LIBRARY";
                    break;
                case 4019:
                    errorDescription += " ERR_CANNOT_CALL_FUNCTION";
                    break;
                case 4020:
                    errorDescription += " ERR_EXTERNAL_CALLS_NOT_ALlowED";
                    break;
                case 4021:
                    errorDescription += " ERR_NO_MEMORY_FOR_RETURNED_STR";
                    break;
                case 4022:
                    errorDescription += " ERR_SYSTEM_BUSY";
                    break;
                case 4051:
                    errorDescription += " ERR_INVALID_FUNCTION_PARAMVALUE";
                    break;
                case 4062:
                    errorDescription += " ERR_STRING_PARAMETER_EXPECTED";
                    break;
                case 4063:
                    errorDescription += " ERR_INTEGER_PARAMETER_EXPECTED";
                    break;
                case 4064:
                    errorDescription += " ERR_DOUBLE_PARAMETER_EXPECTED";
                    break;
                case 4065:
                    errorDescription += " ERR_ARRAY_AS_PARAMETER_EXPECTED";
                    break;
                case 4066:
                    errorDescription += " ERR_HISTORY_WILL_UPDATED";
                    break;
                case 4067:
                    errorDescription += " ERR_TRADE_ERROR";
                    break;
                case 4105:
                    errorDescription += " ERR_NO_ORDER_SELECTED";
                    break;
                case 4106:
                    errorDescription += " ERR_UNKNOWN_SYMBOL";
                    break;
                case 4107:
                    errorDescription += " ERR_INVALID_PRICE_PARAM";
                    break;
                case 4108:
                    errorDescription += " ERR_INVALID_TICKET";
                    break;
                case 4109:
                    errorDescription += " ERR_TRADE_NOT_ALLOWED";
                    break;
                default:
                    errorDescription += " unknown_err_code " + err.ToString();
                    break;
            }
            return errorDescription;
        }

        // Get Total number of Open orders
        internal int GetTotalOrders()
        {
            int total = 0;
            if (!simulate) // Alveo
            {
                total = 0;
                s.buyOpenOrders.Clear();
                s.sellOpenOrders.Clear();
                int count = OrdersTotal();
                for (int pos = 0; pos < count; pos++)
                {
                    if (OrderSelect(pos, SELECT_BY_POS) == false)
                        continue;
                    string cm = OrderComment();
                    if (cm == null || cm.Length < 1)
                        continue;
                    if (!cm.Contains(magicStr))
                        continue;
                    DateTime ctm = OrderCloseTime();
                    if (ctm.Year > 1980)
                        continue;                           // closed
                    var order = new Order();
                    order.Id = OrderTicket();
                    if (order.Id < 0)
                        continue;
                    order.Symbol = symbol;
                    order.Side = (TradeSide)OrderType();
                    order.Comment = OrderComment();
                    order.OpenPrice = (decimal)OrderOpenPrice();
                    order.StopLoss = (decimal)OrderStopLoss();
                    order.TakeProfit = (decimal)OrderTakeProfit();
                    order.OpenDate = OrderOpenTime();
                    order.CloseDate = ctm;
                    order.Quantity = (decimal)OrderLots();
                    order.Type = TradeType.Market;
                    switch (order.Side)
                    {
                        case TradeSide.Buy:
                            s.buyOpenOrders.Add(order.Id, order);
                            break;
                        case TradeSide.Sell:
                            s.sellOpenOrders.Add(order.Id, order);
                            break;
                        default:
                            continue;   // unknown
                    }
                    total++;
                }
            }
            else    // simulate
                total = s.buyOpenOrders.Count + s.sellOpenOrders.Count;
            return total;
        }

        internal double GetProfit(Order order)
        {
            double profit = 0;
            double dPips = 0;
            int ticketID = (int)order.Id;
            if (!simulate)
            {
                if (OrderSelect(ticketID, SELECT_BY_TICKET) == true)
                    profit = OrderProfit();
                else
                    profit = 0;
            }
            else  // simulate
            {
                var side = order.Side;
                dPips = GetPriceDiff(order) * pipPos;
                dPips *= (side == TradeSide.Buy) ? 1.0 : -1.0;
                profit = (double)order.Quantity * dPips * 10;
            }
            return profit;
        }

        // return OrderCloseTime if closed, else return datetime0
        internal datetime GetOrderCloseTime(int ticketID)
        {
            Order order = null;
            datetime closeTime = datetime0;
            if (!simulate)
            {
                if (OrderSelect(ticketID, SELECT_BY_TICKET) == true)
                    closeTime = OrderCloseTime();
                else
                    closeTime = datetime0;
            }
            else // simulate
            {
                try
                {
                    if (s.buyOpenOrders.Count > 0)
                        if (s.buyOpenOrders.Keys.Contains(ticketID))
                            order = s.buyOpenOrders[ticketID];
                }
                catch (KeyNotFoundException) // order not in buyOpenOrders
                {
                    order = null;
                }
                if (order == null)
                {
                    try
                    {
                        if (s.sellOpenOrders.Count > 0)
                            if (s.sellOpenOrders.Keys.Contains(ticketID))
                                order = s.sellOpenOrders[ticketID];
                    }
                    catch (KeyNotFoundException) // order not in sellOpenOrders
                    {
                        order = null;
                    }
                }
                if (order != null) // order found
                {
                    if (order.CloseDate > DateTime.MinValue)  // order.CloseDate not default value
                        closeTime = order.CloseDate;
                    else
                        closeTime = datetime0;
                }
            }
            return closeTime;
        }

        // return OrderClosePrice - OrderOpenPrice
        internal double GetPriceDiff(Order order)
        {
            double priceDiff = 0;
            if (!simulate) // Alveo functions
                priceDiff = OrderClosePrice() - OrderOpenPrice();
            else
                priceDiff = (double)(order.ClosePrice - order.OpenPrice);
            return priceDiff;
        }

        internal double GetBid()
        {
            var digits = GetDigits();
            double bid = 0;
            if (!simulate)
                bid = Bid;
            else
                bid = Math.Round(curPrice - 5 * point, digits);
            return bid;
        }

        internal double GetAsk()
        {
            var digits = GetDigits();
            double ask = 0;
            if (!simulate)
                ask = Ask;
            else
                ask = Math.Round(curPrice + 5 * point, digits);
            return ask;
        }

        // Open a new Market Order in the targetDir with orderSL and orderTP
        public int CreateOrder(double percent = 1.0, int exprTimeMin = 0, double lotsize = 0.01, int stoploss = 100, int takeprofit = 100)
        {
            int cmd;
            double price;
            double orderSL;
            double orderTP;
            var digits = GetDigits();  // Normalize parameters to digits decimal places for Alveo function
            var qty = NormalizeDouble(Math.Max(Math.Abs(percent * lotsize), 0), 2);
            double stoplossLMT = (double)stoploss;
            double takeprofitLMT = (double)takeprofit;
            double ask = GetAsk();
            double bid = GetBid();
            if (!optimize)
                LogPrint("CreateOrder: ask=" + ask.ToString("F5")
                    + " bid=" + bid.ToString("F5"));
            if (s.targetDir > 0)  // Buy
            {
                cmd = OP_BUY;
                price = ask;
                orderSL = (stoplossLMT > 0) ? bid - stoplossLMT * point : 0;
                orderTP = (takeprofitLMT > 0) ? ask + takeprofitLMT * point : 0;
                if (!optimize)
                    LogPrint("CreateOrder: OP_BUY price" + ask
                        + " orderSL=" + orderSL
                        + " orderTP=" + orderTP
                        + " open=" + curTime.ToString()
                        );
            }
            else // Sell
            {
                cmd = OP_SELL;
                price = bid;
                orderSL = (stoplossLMT > 0) ? ask + stoplossLMT * point : 0;
                orderTP = (takeprofitLMT > 0) ? bid - takeprofitLMT * point : 0;
                if (!optimize)
                    LogPrint("CreateOrder: OP_SELL price" + bid
                        + " orderSL=" + orderSL
                        + " orderTP=" + orderTP
                        + " open=" + curTime.ToString()
                        );
            }
            string tf = GetTimeframe().ToString();
            string cmt = strategy + "(" + tf + ")" + " " + symbol;
            int ticket = TryCreateOrder(cmd, qty, price, orderSL, orderTP, exprTimeMin: exprTimeMin, comment: cmt);  // open original order with TP = orderTP
            if (false && ticket > 0 && (cmd == OP_BUY || cmd == OP_SELL))
            {
                var order = new Order();
                order.Id = ticket;
                order.Symbol = symbol;
                order.Side = (cmd == OP_BUY) ? TradeSide.Buy : TradeSide.Sell;
                order.Comment = cmt;
                order.OpenPrice = (decimal)price;
                order.StopLoss = (decimal)orderSL;
                order.TakeProfit = (decimal)orderTP;
                order.OpenDate = curTime;
                order.CloseDate = DateTime.MinValue;
                order.Quantity = (decimal)qty;
                order.Type = TradeType.Market;
                order.Profit = stoploss;
                switch (order.Side)
                {
                    case TradeSide.Buy:
                        s.buyOpenOrders.Add(order.Id, order);
                        break;
                    case TradeSide.Sell:
                        s.sellOpenOrders.Add(order.Id, order);
                        break;
                }
            }
            return ticket;
        }

        // try to create order with the specified parameters
        public int TryCreateOrder(int cmd, double orderQty, double price, double orderSL = 0, double orderTP = 0, string comment = "", int magic = 0, int exprTimeMin = 0, color clr = null)
        {
            var digits = GetDigits();  // Normalize parameters to digits decimal places for Alveo function
            price = NormalizeDouble(price, digits);
            orderSL = NormalizeDouble(orderSL, digits);
            orderTP = NormalizeDouble(orderTP, digits);
            var qty = NormalizeDouble(Math.Abs(orderQty), 2);
            int ticket = -1;  // dummy ticket number
            s.ordersThisHour++;  // increment ordersThisHour count
            ticket = SendOrder(cmd, qty, price, 5, orderSL, orderTP, comment: comment, expr: exprTimeMin);
            if (!optimize)
            {
                string msg = "TryCreateOrder: Order created. Id=" + ticket + " Qty=" + qty + " Side=" + cmd + " price=" + price + " SL=" + orderSL + " TP=" + orderTP;
                LogPrint(msg);
            }
            s.stats.numOrders++;  // increment numTrades count
            return ticket;
        }

        // Get specified MarketInfo Item
        public double GetMarketInfo(string symbol, int item)
        {
            switch (item)
            {
                case MODE_BID:
                    return (simulate) ?
                        curPrice - 4 * point
                        : Bid;  // Alveo value
                case MODE_ASK:
                    return (simulate) ?
                        curPrice + 4 * point
                        : Ask; // Alveo value
                case MODE_STOPLEVEL:
                    return (simulate) ?
                        0
                        : GetMarketInfo(symbol, MODE_STOPLEVEL); // Alveo value
                default:
                    return 0; // unknown item
            }
        }

        // Get Normalize Digits from Alveo
        internal int GetDigits()
        {
            if (!simulate)
                return Digits;  // Alveo function
            else if (symbol.EndsWith("JPY"))
                return 3;
            return 5;
        }

        // Send OrderSend request to Alveo
        internal int SendOrder(int cmd, double volume, double price, int slippage = 5, double stoploss = 0, double takeprofit = 0, string comment = "", int magic = 0, int expr = 0, color clr = null)
        {
            if (price <= 0)
                throw new Exception("SendOrder: invalid price !!");
            int ticket = -1;
            var exprDT = GetCurTime() + TimeSpan.FromMinutes((double)expr + 1);
            if (!simulate)  // Alveo functions
            {
                if (comment.Length == 0)
                    comment = strategy;
                int err = 0;
                int cnt = 0;
                var digits = GetDigits();
                var thePrice = NormalizeDouble(price, digits);
                while (ticket < 0 && err != 4109 && err != 132 && err != 133)
                {
                    if (expr > 0)
                    {
                        LogPrint(" curPrice=" + curPrice);
                        LogPrint(" SendOrder: cmd=" + cmd
                            + " vol=" + volume
                            + " price=" + thePrice
                            + " slippage=" + slippage
                            + " stoploss=" + stoploss
                            + " takeprofit=" + takeprofit
                            + " exprDT=" + exprDT.ToLongTimeString()
                            );
                        ticket = OrderSend(symbol, cmd, volume, thePrice, slippage, stoploss, takeprofit, magicStr, 0, exprDT);
                    }
                    else
                    {
                        ticket = OrderSend(symbol, cmd, volume, thePrice, slippage, stoploss, takeprofit, magicStr);
                        LogPrint(" SendOrder: cmd=" + cmd
                            + " vol=" + volume
                            + " price=" + price
                            + " slippage=" + slippage
                            + " stoploss=" + stoploss
                            + " takeprofit=" + takeprofit
                            );
                    }
                    if (ticket < 0) // Alveo returned err
                    {
                        err = GetLastError();
                        if (err == 4108)
                            LogPrint("about line 2646 OrderSend failed !!");
                        cnt++;
                        if (cnt > 10)  // try up to 10 times before Exception
                        {
                            string msg = "SendOrder: OrderSend failed !! err=" + err.ToString();
                            LogPrint(msg);
                            throw new Exception(msg);
                        }
                    }
                    if (!simulate)
                        Sleep(2000); // Sleep 2 sec between OrderSends
                }
            }
            else // simulate, Simulator, not Alveo
            {
                ticket = ++ticketNum;  // update simulated ticketNum
            }
            if (ticket >= 0)  // ticket valid, create new Order
            {
                Order order = new Order();
                order.Id = ticket;
                order.Symbol = symbol;
                order.Quantity = (decimal)volume;
                order.StopLoss = (decimal)stoploss;
                order.TakeProfit = (decimal)takeprofit;
                order.CloseDate = DateTime.MinValue;
                order.ExpirationDate = exprDT;
                order.OpenDate = GetCurTime();
                order.CloseType = CloseType.Unknown;
                if (cmd == OP_BUY)
                {
                    order.Type = TradeType.Market;
                    order.Side = TradeSide.Buy;
                    order.OpenPrice = (decimal)price;
                    s.buyOpenOrders.Add(ticket, order);  // add to buyOpenOrders list
                }
                else if (cmd == OP_SELL)
                {
                    order.Type = TradeType.Market;
                    order.Side = TradeSide.Sell;
                    order.OpenPrice = (decimal)price;
                    s.sellOpenOrders.Add(ticket, order);  // add to sellOpenOrders list
                }
                if (!optimize) // save Order data in order.Comment and LogPrint info
                {
                    string msg = String.Format(",{0},{1},{2},{3},{4},{5},{6}",
                        (comment.Length > 0) ? comment : ((cmd == OP_BUY) ? "B" : "S"),
                        order.Id.ToString(),
                        order.Quantity.ToString(),
                        order.OpenDate.ToShortDateString() + " " + order.OpenDate.ToLongTimeString(),
                        order.OpenPrice.ToString("F5"),
                        order.StopLoss.ToString("F5"),
                        order.TakeProfit.ToString("F5"));
                    order.Comment = magicStr + msg;
                    LogPrint("SendOrder: Opened order=" + order.Id
                        + " OpenPrice=" + order.OpenPrice.ToString("F5")
                        + " vol=" + order.Quantity.ToString("F5")
                        + " TakeProfit=" + order.TakeProfit.ToString("F5"));
                }
            }
            return ticket;  // unique ticket number
        }

        internal bool ModifyOrder(int ticket, int stoploss, ref Order order)
        {
            bool res = false;
            int err = 0;
            int cnt = 0;
            if (!simulate)  // Alveo functions
            {
                if (!OrderSelect(ticket, SELECT_BY_TICKET))
                    return false;
                DateTime dt = OrderCloseTime();
                if (dt.Year > 1970)
                    return false;  // already closed
                double mult = (OrderType() == OP_BUY) ? 1 : (OrderType() == OP_SELL) ? -1 : 0;
                var dist = mult * (curPrice - OrderStopLoss());
                if (dist - stoploss > 10 * point)
                {
                    while (!res && err != 4109 && err != 132 && err != 133)
                    {
                        var prc = OrderOpenPrice();
                        var orderSL = (curPrice - mult * stoploss);
                        var sl = NormalizeDouble(orderSL, Digits);
                        var tp = OrderTakeProfit();
                        res = OrderModify(ticket, prc, sl, tp, 0);
                        if (!res) // Alveo returned err
                        {
                            err = GetLastError();
                            if (!optimize && err == 4108)
                                LogPrint("about line 280 OrderSend failed !!");
                            cnt++;
                            if (cnt > 10)  // try up to 10 times before Exception
                                throw new Exception("SendOrder: OrderModify failed !! err=" + err.ToString());
                        }
                        else // res = true
                        {
                            if (!optimize)
                                LogPrint(" ModifyOrder: ticket=" + ticket + " stoploss=" + stoploss + " OrderOpenPrice=" + prc + " SLprice=" + sl);
                        }
                        if (!simulate)
                            Sleep(200); // Sleep 0.2 sec between OrderModify
                    }
                }
                else  // no need to change stoploss
                    res = false;
            }
            else // simulate
            {
                double mult = (order.Side == TradeSide.Buy) ? 1 : -1;
                double diff = mult * (curPrice - (double)order.StopLoss);
                if (diff > (double)(order.StopLoss + 10) * point)
                {
                    var digits = GetDigits();
                    var orderSL = NormalizeDouble(curPrice - mult * stoploss * point, digits);
                    order.StopLoss = (decimal)orderSL;
                }
            }
            return res;  // unique ticket number
        }

        // emulate Alveo Chart.TimeFrame
        internal TimeFrame GetTimeframe()
        {
            TimeFrame tf = TimeFrame.M5;
            if (!simulate)
                tf = Chart.TimeFrame;
            else
                tf = simTimeframe;
            return tf;
        }

        internal double GetOrderClosePrice(int orderID, Order order)
        {
            double clsPrice = double.MinValue;
            if (!simulate)
            {
                if (OrderSelect(orderID, SELECT_BY_TICKET) == true)
                {
                    clsPrice = OrderClosePrice();
                }
                else
                    if (!optimize) LogPrint("GetOrderClosePrice: OrderSelect() returned error - " + GetLastError());
            }
            else
            {
                return (double)order.ClosePrice;
            }
            return clsPrice;
        }

        internal int GetAccountNumber()  // Get account number
        {
            int account = 0;                    // default value for simulation
            if (!simulate)                      // Alveo running
                account = AccountNumber();      // Alveo function
            return account;
        }
        #endregion

        internal class iRSIobj
        {
            EMAobj avgGain;
            EMAobj avgLoss;
            double Period;
            double Threshold;
            double prevPrice;
            internal double value;
            internal double prevValue1;
            internal double prevValue2;
            internal bool crossFromBelow;
            internal bool crossFromAbove;

            internal iRSIobj()
            {
                Period = 14;
                Threshold = 1 * 1e-6;
                prevPrice = 0;
                value = double.MinValue;
                prevValue1 = value;
                prevValue2 = value;
                crossFromBelow = false;
                crossFromAbove = false;
            }

            internal iRSIobj(int period, int threshold = 0) : this()   // do iRSI() first
            {
                Period = period;
                Threshold = (double)threshold * 1e-6;
                prevPrice = 0;
                avgGain = new EMAobj(period, threshold);
                avgLoss = new EMAobj(period, threshold);
            }

            internal void Init(double thePrice)  // Initialize Indicator
            {
                prevPrice = thePrice;
                avgGain.Init(0);
                avgLoss.Init(0);
                value = 50;
                prevValue1 = value;
                prevValue2 = value;
                crossFromBelow = false;
                crossFromAbove = false;
            }

            internal double Calc(double thePrice)  // Calculate Indicator values
            {
                prevValue2 = prevValue1;
                prevValue1 = value;
                // iRSIval = 100 - 100 / (1 + avgGain.value/ avgLoss.value)
                if (Period < 1)
                    throw new Exception("iRSIcalc: period < 1, invalid !!");
                if (Threshold < 0)
                    throw new Exception("iRSIcalc: Threshold < 1e-10, invalid !!");
                var diff = thePrice - prevPrice;
                prevPrice = thePrice;
                avgGain.Calc((diff > 0) ? diff : 0);
                avgLoss.Calc((diff < 0) ? -diff : 0);
                var ratio = Math.Max(avgGain.value, 0) / Math.Max(avgLoss.value, 10e-20);
                if (ea.simulate)
                    value = 100 - 100 / (1 + ratio);
                else
                    value = ea.iRSI(null, 0, (int)Math.Round(Period), PRICE_MEDIAN, 1);
                crossFromBelow = (value > 50 && (prevValue1 < 50 || prevValue2 < 50));
                crossFromAbove = (value < 50 && (prevValue1 > 50 || prevValue2 > 50));
                return value;
            }
        }

        internal class EMAobj
        {
            int Period;
            double Threshold;
            internal double value;
            internal double prevValue;
            internal bool isRrising;
            internal bool isFalling;
            internal int trendDir;
            internal int prevTrendDir;
            internal bool trendChanged;
            double K;

            // EMAobj constructor
            internal EMAobj()
            {
                Period = 12;
                Threshold = 1 * 1e-6;
                isRrising = false;
                isFalling = false;
                trendDir = 0;
                prevTrendDir = 0;
                trendChanged = false;
                value = double.MinValue;
                prevValue = value;
                K = 1.0;
            }

            // EMAobj constructor with input parameters
            internal EMAobj(int period, double threshold = 0) : this()  // call EMAobj() first
            {
                Period = period;
                Threshold = (double)threshold * 1e-6;
                K = 2.0 / (Period + 1.0);
            }

            // Initialize EMAobj
            internal void Init(double thePrice)
            {
                isRrising = false;
                isFalling = false;
                trendDir = 0;
                prevTrendDir = 0;
                trendChanged = false;
                prevValue = thePrice;
                value = thePrice;
            }

            // Calculate EMAobj;
            internal double Calc(double thePrice)
            {
                if (Period < 1)
                    throw new Exception("EMA: period < 1, invalid !!");
                if (Threshold < 0)
                    throw new Exception("EMA: Threshold < 0, invalid !!");
                if (prevValue == double.MinValue)
                    Init(thePrice);
                //EMA = (Price[today] x K) + (EMA[prev] x (1 – K)); K = 2 / (N + 1)
                prevValue = value;
                value = (thePrice * K) + prevValue * (1 - K);
                var diff = value - prevValue;
                isRrising = (diff > Threshold);
                isFalling = (diff < -Threshold);
                if (trendDir != 0)
                    prevTrendDir = trendDir;
                trendDir = isRrising ? 1 : (isFalling ? -1 : 0);
                trendChanged = (trendDir * prevTrendDir < 0);
                return value;
            }
        }
    }
}